/*
*  Name:      TBaseConnectionManager.java
*  Author:    Rafael Jesus Alcantara Perez
*  Summary:   Class for handling connections in JAVA.
*  Date:      $Date: 2002/08/19 12:59:57 $
*  Revision:  $Revision: 1.1 $
*
*  Copyright (C) 2002  Rafael Jesus Alcantara Perez <rafa@dedalo-ing.com>
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
*  MA 02111-1307, USA.
*/

///  Name-space for UESQL.
package org.uesqlc;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;


//   ///  Unary function object to delete parameters.
//   template <typename TItem>
//   struct MWriteDebugInfo : public std::unary_function<TItem, void>
//   {
// 
//       void operator () (TItem* ptITEM) const
//       {
// #       ifdef _UESQL_DEBUG__
//         mpcl::tLog << ", " << ptITEM->get();
// #       endif  // not _UESQL_DEBUG__
//       }
// 
//   }  // struct MWriteDebugInfo

///  UESQL base manager class.
public abstract class TBaseConnectionManager extends TSqlstateDescriptor
{

  ///  Map that binds connection names to connections.
  protected Hashtable   tConnectionHashtable;

  ///  Current connection (from \a tConnectionHashtable).
  protected Connection   tCurrentConnection;

  ///  Input embedded variable string values queue.
  protected Vector   tInputEmbeddedVariableVector;

  ///  Output embedded variable string pointers vector.
  protected Vector   tOutputEmbeddedVariableVector;

  ///  String with the identifier of the current connection.
  protected String   yCurrentConnection;


  //
  //  C O N S T R U C T O R S
  //

  ///  Builds a new instance.
  public TBaseConnectionManager()
  {
    tConnectionHashtable          = new Hashtable();
    tCurrentConnection            = null;
    tInputEmbeddedVariableVector  = new Vector();
    tOutputEmbeddedVariableVector = new Vector();
    yCurrentConnection            = null;
    initialize();
  }

  ///  Destroys the instance.
  public void terminate() throws SQLException
  {
    finish();
  }

  /**
  *  Binds in connection map, connection name \a kyCONNECTION_NAME to
  *  \a tCONNECTION.
  *  @param kyCONNECTION_NAME Connection name.
  *  @param tCONNECTION       Connection.
  */
  public void bind (final String kyCONNECTION_NAME, Connection tCONNECTION)
    throws SQLException
  {
    tConnectionHashtable.put (kyCONNECTION_NAME, tCONNECTION);
    tCONNECTION.setAutoCommit (false);
  }

  /**
  *  Commits current connection.
  *  @pre isConnected()
  */
  public void commitCurrent()
  {
    try
    {
      tCurrentConnection.commit();
    }
    catch (final SQLException ktEXCEPTION)
    {
      diagnose (ktEXCEPTION);
    }
  }

  /**
  *  Diagnoses last SQLException thrown. It updates the SQL status variables.
  *  @param ktEXCEPTION Source exception.
  */
  public abstract void diagnose (final SQLException ktEXCEPTION);

//   /**
//   *  Diagnoses  statement result.  It only checks  the result of the  statement
//   *  executed. This member function (\a diagnose()) updates properly SQL status
//   *  variables.
//   *  @param ktCONNECTION Connection where the last statement was executed.
//   */
//   public abstract void diagnose (final Connection ktCONNECTION);

  /**
  *  Unbinds connection named \a kyCONNECTION_NAME.
  *  @param kyCONNECTION_NAME Source connection name.
  *  @warning kyCONNECTION_NAME  mustn't  be  passed  as reference to avoid
  *           problems  about passing  \a yCurrentConnection  as  argument.
  */
  public void unbind (final String kyCONNECTION_NAME)
  {
    Connection   tConnection = (Connection) tConnectionHashtable.get (kyCONNECTION_NAME);

    tConnectionHashtable.remove (kyCONNECTION_NAME);
    try
    {
      tConnection.close();
    }
    catch (final SQLException ktEXCEPTION)
    {
      diagnose (ktEXCEPTION);
    }
    if ( kyCONNECTION_NAME == yCurrentConnection )
    {
      yCurrentConnection = null;
    }
  }

  ///  Unbinds current connection (if any).
  public void unbindCurrent()
  {
    if ( isConnected() )
    {
      unbind (yCurrentConnection);
    }
  }

  ///  Unbinds all connections.
  public void unbindAll()
  {
    Enumeration   tElementEnumeration = tConnectionHashtable.keys();

    while ( tElementEnumeration.hasMoreElements() )
    {
      unbind ((String) tElementEnumeration.nextElement());
    }
  }

  /**
  *  Clears  embedded  variable  containers.  It  must  be  called  at  the
  *  beginning of each new instruction.
  */
  public void clearQueues()
  {
    tInputEmbeddedVariableVector.clear();
    tOutputEmbeddedVariableVector.clear();
  }

  /**
  *  Executes the statement \a kySTATEMENT instantiating the input embedded
  *  variables (tagged with '$z') in the statement \a tSTATEMENT.
  *  @param tSTATEMENT  Statement where the string is executed.
  *  @param kySTATEMENT String with the source statement.
  */
  public ResultSet execute (final String kySTATEMENT)
  {

    Statement   tStatement;
    ResultSet   tResultSet = null;

    try
    {
      tStatement = createStatement();
      tResultSet = tStatement.executeQuery (instantiate (kySTATEMENT));
    }
    catch (final SQLException ktEXCEPTION)
    {
      diagnose (ktEXCEPTION);
    }
    return tResultSet;

  }  // execute()

  /**
  *  Executes the statement kySTATEMENT instantiating
  *  the input embedded variables (tagged with '$z'). The result
  *  of the execution is omitted.
  *  @param kySTATEMENT String with the source statement.
  */
  public void executeClean (final String kySTATEMENT)
    throws SQLException
  {
    Statement   tStatement;

    try
    {
      tStatement = createStatement();
      tStatement.execute (instantiate (kySTATEMENT));
      tStatement.close();
    }
    catch (final SQLException ktEXCEPTION) {}
  }

  /**
  *  Executes  the  statement \a kySTATEMENT  instantiating the  input embedded
  *  variables (tagged with '$z') in the statement \a tSTATEMENT. The statement
  *  can not return more than one tuple.
  *  @param tSTATEMENT  Statement where the string is executed.
  *  @param kySTATEMENT String with the source statement.
  */
  public ResultSet executeSingleRow (final String kySTATEMENT)
  {

    Statement   tStatement;
    ResultSet   tResultSet = null;

    try
    {
      tStatement = createStatement();
      tResultSet = tStatement.executeQuery (instantiate (kySTATEMENT));
    }
    catch (final SQLException ktEXCEPTION)
    {
      diagnose (ktEXCEPTION);
    }
    if ( TSqlstate.SQLCODE == 0 )
    {
      try
      {
        if ( tResultSet.next() && tResultSet.next() )
        {
          // 
          //  Condition:  cardinality violation
          // 
          TSqlstate.SQLCODE    = -1;
          TSqlstate.SQLMESSAGE = "";
          TSqlstate.SQLSTATE   = "21000";
        }
        tResultSet.beforeFirst();
      }
      catch (final SQLException ktEXCEPTION)
      {
        diagnose (ktEXCEPTION);
      }
    }
    return tResultSet;

  }  // executeSingleRow()

  ///  Finishes all managed issues.
  public void finish() throws SQLException
  {
    unbindAll();
    clearQueues();
    tConnectionHashtable.clear();
    yCurrentConnection = null;
  }

  /**
  *  Gets a pointer to the SQL data type of the output embedded variable at
  *  position kiINDEX in the  output embedded variable vector.  It  must be
  *  called for writing results on output embedded variables.
  *  @param kiINDEX Position of the variable.
  *  @return Pointer to the output embedded variable.
  */
  public IType getFromOutputQueue (final int kiINDEX)
  {
//     if ( zINDEX >= tOutputEmbeddedVariableVector.size() )
//     {
//       throw mpcl::TExceptionEL_Constraint ("out of range");
//     }
    return (IType) tOutputEmbeddedVariableVector.get (kiINDEX);
  }

  ///  Initializes the default connection.
  public abstract void initialize();

  /**
  *  Returns a string that is result of instanting the input
  *  embedded variables (if any).
  *  @param kySTATEMENT Source statement.
  *  @return The instantiated string.
  */
  public String instantiate (final String kySTATEMENT)
  {

    int            iPatternIndex;
    StringBuffer   yStatementInstance = new StringBuffer (kySTATEMENT);
    IType          tType;

    while ( !tInputEmbeddedVariableVector.isEmpty() )
    {
      iPatternIndex = yStatementInstance.indexOf ("$z");
      if ( -1 == iPatternIndex )
      {
        break;
      }
      else
      {
        tType = (IType) tInputEmbeddedVariableVector.firstElement();
        yStatementInstance.replace (iPatternIndex, iPatternIndex + 2, tType.toString());
        tInputEmbeddedVariableVector.remove (0);
      }
    }
    return yStatementInstance.toString();

  }  // instantiate()

  /**
  *  Pushes on the input embedded variable queue the variable
  *  value \a ktEMBEDDED_VALUE.
  *  @param ktEMBEDDED_VALUE Input embedded variable value.
  */
  public void pushOnInputQueue (final IType ktEMBEDDED_VALUE)
  {
    tInputEmbeddedVariableVector.add (ktEMBEDDED_VALUE.toString());
  }

  /**
  *  Pushes on the output embedded variable vector the variable
  *  reference \a tSQL_DATA_TYPE.
  *  @param tSQL_DATA_TYPE Output embedded variable reference.
  */
  public void pushOnOutputQueue (IType tSQL_DATA_TYPE)
  {
    tOutputEmbeddedVariableVector.add (tSQL_DATA_TYPE);
  }

  ///  Resets SQL status variables.
  public void resetStatus()
  {
    //
    //  Condition:  successful completion
    //
    TSqlstate.SQLCODE    = 0;
    TSqlstate.SQLMESSAGE = "";
    TSqlstate.SQLSTATE   = "00000";
  }

  /**
  *  Sets current connection to connection named kyCONNECTION_NAME.
  *  @param kyCONNECTION_NAME Source connection name.
  */
  public void setCurrentConnection (final String kyCONNECTION_NAME)
  {

    if ( tConnectionHashtable.contains (kyCONNECTION_NAME) )
    {
      yCurrentConnection = kyCONNECTION_NAME;
      tCurrentConnection = (Connection) tConnectionHashtable.get (kyCONNECTION_NAME);
// #         ifdef _UESQL_DEBUG__
//           mpcl::tLog << mpcl::header << "SET CONNECTION '" << kyCONNECTION_NAME << "';\n";
// #         endif  // not _UESQL_DEBUG__
      resetStatus();
    }
    else
    {
      //
      //  Condition:           connection exception
      //  Class subcondition:  connection does not exist
      //
      TSqlstate.SQLCODE    = -1;
      TSqlstate.SQLMESSAGE = "";
      TSqlstate.SQLSTATE   = "08003";
    }
      
  }  // setCurrentConnection()

//   /**
//   *  Updates output variables  (which  have  been updated passively using a
//   *  temporary buffer).
//   */
//   public void updateOutputVariables()
//   {
//     for (int I = 0; ( I < tOutputEmbeddedVariableVector.size() ) ;++I)
//     {
//       tOutputEmbeddedVariableVector.get (I).updateFromBuffer();
//     }
// // # ifdef _UESQL_DEBUG__
// //   mpcl::tLog << mpcl::header << "FETCHED (";
// //   if ( !tOutputEmbeddedVariableVector.empty() )
// //   {
// //     mpcl::tLog << tOutputEmbeddedVariableVector.front()->get();
// //     std::for_each ( tOutputEmbeddedVariableVector.begin() + 1 ,
// //                     tOutputEmbeddedVariableVector.end()       ,
// //                     MWriteDebugInfo<IType>()     );
// //   }
// //   mpcl::tLog << ");\n";
// // # endif  // not _UESQL_DEBUG__
//   }  // updateOutputVariables()


  //
  //  S E L E C T O R S
  //

  /**
  *  Creates a new statement for the current connection.
  *  @return The newly created statement.
  */
  protected Statement createStatement() throws SQLException
  {
    return tCurrentConnection.createStatement();
  }

  /**
  *  Returns the connection name for the current connection.
  *  @return The connection name for the current connection.
  */
  public final String currentConnectionName()
  {
    return yCurrentConnection;
  }

  /**
  *  Returns true if there is a default connection.
  *  @return True if there is a default connection.
  */
  public boolean isConnected()
  {
    return ( yCurrentConnection != null );
  }

  /**
  *  Returns true if there is an active connection.
  *  @param kyCONNECTION_NAME Source connection name to check.
  *  @return True if there is an active connection.
  */
  public boolean isConnected (final String kyCONNECTION_NAME)
  {
    return tConnectionHashtable.contains (kyCONNECTION_NAME);
  }

  /**
  *  Checks if a NOT FOUND condition has happened.
  *  @return True if it has happened.
  */
  public boolean isNotFoundCondition()
  {
    return ( TSqlstate.SQLSTATE == TSqlstate.no_data._kyNO_SUBCLASS );
  }

  /**
  *  Checks if a SQLERROR condition has happened.
  *  @return True if it has happened.
  */
  public boolean isSqlerrorCondition()
  {
    return ( TSqlstate.SQLCODE != 0 );
  }

}  // class TBaseConnectionManager
